package io.quarkus.registry.catalog.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import io.quarkus.registry.model.Release;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link Extension}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableExtension.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableExtension implements Extension {
  private final String groupId;
  private final String artifactId;
  private final List<Release> releases;

  private ImmutableExtension(
      String groupId,
      String artifactId,
      List<Release> releases) {
    this.groupId = groupId;
    this.artifactId = artifactId;
    this.releases = releases;
  }

  /**
   * @return The value of the {@code groupId} attribute
   */
  @JsonProperty("group-id")
  @Override
  public String getGroupId() {
    return groupId;
  }

  /**
   * @return The value of the {@code artifactId} attribute
   */
  @JsonProperty("artifact-id")
  @Override
  public String getArtifactId() {
    return artifactId;
  }

  /**
   * @return The value of the {@code releases} attribute
   */
  @JsonProperty("releases")
  @Override
  public List<Release> getReleases() {
    return releases;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getGroupId() groupId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for groupId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withGroupId(String value) {
    String newValue = Objects.requireNonNull(value, "groupId");
    if (this.groupId.equals(newValue)) return this;
    return new ImmutableExtension(newValue, this.artifactId, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getArtifactId() artifactId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for artifactId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withArtifactId(String value) {
    String newValue = Objects.requireNonNull(value, "artifactId");
    if (this.artifactId.equals(newValue)) return this;
    return new ImmutableExtension(this.groupId, newValue, this.releases);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getReleases() releases}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withReleases(Release... elements) {
    List<Release> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableExtension(this.groupId, this.artifactId, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getReleases() releases}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of releases elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withReleases(Iterable<? extends Release> elements) {
    if (this.releases == elements) return this;
    List<Release> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return new ImmutableExtension(this.groupId, this.artifactId, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtension} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtension
        && equalTo((ImmutableExtension) another);
  }

  private boolean equalTo(ImmutableExtension another) {
    return groupId.equals(another.groupId)
        && artifactId.equals(another.artifactId);
  }

  /**
   * Computes a hash code from attributes: {@code groupId}, {@code artifactId}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + groupId.hashCode();
    h += (h << 5) + artifactId.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Extension} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Extension{"
        + "groupId=" + groupId
        + ", artifactId=" + artifactId
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Extension {
    String groupId;
    String artifactId;
    List<Release> releases = Collections.emptyList();
    @JsonProperty("group-id")
    public void setGroupId(String groupId) {
      this.groupId = groupId;
    }
    @JsonProperty("artifact-id")
    public void setArtifactId(String artifactId) {
      this.artifactId = artifactId;
    }
    @JsonProperty("releases")
    public void setReleases(List<Release> releases) {
      this.releases = releases;
    }
    @Override
    public String getGroupId() { throw new UnsupportedOperationException(); }
    @Override
    public String getArtifactId() { throw new UnsupportedOperationException(); }
    @Override
    public List<Release> getReleases() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableExtension fromJson(Json json) {
    ImmutableExtension.Builder builder = ImmutableExtension.builder();
    if (json.groupId != null) {
      builder.groupId(json.groupId);
    }
    if (json.artifactId != null) {
      builder.artifactId(json.artifactId);
    }
    if (json.releases != null) {
      builder.addAllReleases(json.releases);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Extension} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Extension instance
   */
  public static ImmutableExtension copyOf(Extension instance) {
    if (instance instanceof ImmutableExtension) {
      return (ImmutableExtension) instance;
    }
    return ImmutableExtension.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableExtension ImmutableExtension}.
   * <pre>
   * ImmutableExtension.builder()
   *    .groupId(String) // required {@link Extension#getGroupId() groupId}
   *    .artifactId(String) // required {@link Extension#getArtifactId() artifactId}
   *    .addReleases|addAllReleases(io.quarkus.registry.model.Release) // {@link Extension#getReleases() releases} elements
   *    .build();
   * </pre>
   * @return A new ImmutableExtension builder
   */
  public static ImmutableExtension.Builder builder() {
    return new ImmutableExtension.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableExtension ImmutableExtension}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_GROUP_ID = 0x1L;
    private static final long INIT_BIT_ARTIFACT_ID = 0x2L;
    private long initBits = 0x3L;

    private String groupId;
    private String artifactId;
    private List<Release> releases = new ArrayList<Release>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Extension} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Extension instance) {
      Objects.requireNonNull(instance, "instance");
      groupId(instance.getGroupId());
      artifactId(instance.getArtifactId());
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Initializes the value for the {@link Extension#getGroupId() groupId} attribute.
     * @param groupId The value for groupId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("group-id")
    public final Builder groupId(String groupId) {
      this.groupId = Objects.requireNonNull(groupId, "groupId");
      initBits &= ~INIT_BIT_GROUP_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link Extension#getArtifactId() artifactId} attribute.
     * @param artifactId The value for artifactId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("artifact-id")
    public final Builder artifactId(String artifactId) {
      this.artifactId = Objects.requireNonNull(artifactId, "artifactId");
      initBits &= ~INIT_BIT_ARTIFACT_ID;
      return this;
    }

    /**
     * Adds one element to {@link Extension#getReleases() releases} list.
     * @param element A releases element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release element) {
      this.releases.add(Objects.requireNonNull(element, "releases element"));
      return this;
    }

    /**
     * Adds elements to {@link Extension#getReleases() releases} list.
     * @param elements An array of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release... elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Extension#getReleases() releases} list.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("releases")
    public final Builder releases(Iterable<? extends Release> elements) {
      this.releases.clear();
      return addAllReleases(elements);
    }

    /**
     * Adds elements to {@link Extension#getReleases() releases} list.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllReleases(Iterable<? extends Release> elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutableExtension ImmutableExtension}.
     * @return An immutable instance of Extension
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableExtension build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableExtension(groupId, artifactId, createUnmodifiableList(true, releases));
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_GROUP_ID) != 0) attributes.add("groupId");
      if ((initBits & INIT_BIT_ARTIFACT_ID) != 0) attributes.add("artifactId");
      return "Cannot build Extension, some of required attributes are not set " + attributes;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
