package io.quarkus.registry.catalog.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import io.quarkus.registry.model.Release;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link Platform}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutablePlatform.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutablePlatform implements Platform {
  private final String groupId;
  private final String groupIdJson;
  private final String artifactId;
  private final String artifactIdJson;
  private final List<Release> releases;

  private ImmutablePlatform(ImmutablePlatform.Builder builder) {
    this.groupId = builder.groupId;
    this.artifactId = builder.artifactId;
    this.releases = createUnmodifiableList(true, builder.releases);
    if (builder.groupIdJson != null) {
      initShim.groupIdJson(builder.groupIdJson);
    }
    if (builder.artifactIdJson != null) {
      initShim.artifactIdJson(builder.artifactIdJson);
    }
    this.groupIdJson = initShim.getGroupIdJson();
    this.artifactIdJson = initShim.getArtifactIdJson();
    this.initShim = null;
  }

  private ImmutablePlatform(
      String groupId,
      String groupIdJson,
      String artifactId,
      String artifactIdJson,
      List<Release> releases) {
    this.groupId = groupId;
    this.groupIdJson = groupIdJson;
    this.artifactId = artifactId;
    this.artifactIdJson = artifactIdJson;
    this.releases = releases;
    this.initShim = null;
  }

  private static final byte STAGE_INITIALIZING = -1;
  private static final byte STAGE_UNINITIALIZED = 0;
  private static final byte STAGE_INITIALIZED = 1;
  private transient volatile InitShim initShim = new InitShim();

  private final class InitShim {
    private byte groupIdJsonBuildStage = STAGE_UNINITIALIZED;
    private String groupIdJson;

    String getGroupIdJson() {
      if (groupIdJsonBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (groupIdJsonBuildStage == STAGE_UNINITIALIZED) {
        groupIdJsonBuildStage = STAGE_INITIALIZING;
        this.groupIdJson = Objects.requireNonNull(getGroupIdJsonInitialize(), "groupIdJson");
        groupIdJsonBuildStage = STAGE_INITIALIZED;
      }
      return this.groupIdJson;
    }

    void groupIdJson(String groupIdJson) {
      this.groupIdJson = groupIdJson;
      groupIdJsonBuildStage = STAGE_INITIALIZED;
    }

    private byte artifactIdJsonBuildStage = STAGE_UNINITIALIZED;
    private String artifactIdJson;

    String getArtifactIdJson() {
      if (artifactIdJsonBuildStage == STAGE_INITIALIZING) throw new IllegalStateException(formatInitCycleMessage());
      if (artifactIdJsonBuildStage == STAGE_UNINITIALIZED) {
        artifactIdJsonBuildStage = STAGE_INITIALIZING;
        this.artifactIdJson = Objects.requireNonNull(getArtifactIdJsonInitialize(), "artifactIdJson");
        artifactIdJsonBuildStage = STAGE_INITIALIZED;
      }
      return this.artifactIdJson;
    }

    void artifactIdJson(String artifactIdJson) {
      this.artifactIdJson = artifactIdJson;
      artifactIdJsonBuildStage = STAGE_INITIALIZED;
    }

    private String formatInitCycleMessage() {
      List<String> attributes = new ArrayList<>();
      if (groupIdJsonBuildStage == STAGE_INITIALIZING) attributes.add("groupIdJson");
      if (artifactIdJsonBuildStage == STAGE_INITIALIZING) attributes.add("artifactIdJson");
      return "Cannot build Platform, attribute initializers form cycle " + attributes;
    }
  }

  private String getGroupIdJsonInitialize() {
    return Platform.super.getGroupIdJson();
  }

  private String getArtifactIdJsonInitialize() {
    return Platform.super.getArtifactIdJson();
  }

  /**
   * @return The value of the {@code groupId} attribute
   */
  @JsonProperty("group-id")
  @Override
  public String getGroupId() {
    return groupId;
  }

  /**
   * @return The value of the {@code groupIdJson} attribute
   */
  @JsonProperty("group-id-json")
  @Override
  public String getGroupIdJson() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getGroupIdJson()
        : this.groupIdJson;
  }

  /**
   * @return The value of the {@code artifactId} attribute
   */
  @JsonProperty("artifact-id")
  @Override
  public String getArtifactId() {
    return artifactId;
  }

  /**
   * @return The value of the {@code artifactIdJson} attribute
   */
  @JsonProperty("artifact-id-json")
  @Override
  public String getArtifactIdJson() {
    InitShim shim = this.initShim;
    return shim != null
        ? shim.getArtifactIdJson()
        : this.artifactIdJson;
  }

  /**
   * @return The value of the {@code releases} attribute
   */
  @JsonProperty("releases")
  @Override
  public List<Release> getReleases() {
    return releases;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Platform#getGroupId() groupId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for groupId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutablePlatform withGroupId(String value) {
    String newValue = Objects.requireNonNull(value, "groupId");
    if (this.groupId.equals(newValue)) return this;
    return new ImmutablePlatform(newValue, this.groupIdJson, this.artifactId, this.artifactIdJson, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Platform#getGroupIdJson() groupIdJson} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for groupIdJson
   * @return A modified copy of the {@code this} object
   */
  public final ImmutablePlatform withGroupIdJson(String value) {
    String newValue = Objects.requireNonNull(value, "groupIdJson");
    if (this.groupIdJson.equals(newValue)) return this;
    return new ImmutablePlatform(this.groupId, newValue, this.artifactId, this.artifactIdJson, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Platform#getArtifactId() artifactId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for artifactId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutablePlatform withArtifactId(String value) {
    String newValue = Objects.requireNonNull(value, "artifactId");
    if (this.artifactId.equals(newValue)) return this;
    return new ImmutablePlatform(this.groupId, this.groupIdJson, newValue, this.artifactIdJson, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Platform#getArtifactIdJson() artifactIdJson} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for artifactIdJson
   * @return A modified copy of the {@code this} object
   */
  public final ImmutablePlatform withArtifactIdJson(String value) {
    String newValue = Objects.requireNonNull(value, "artifactIdJson");
    if (this.artifactIdJson.equals(newValue)) return this;
    return new ImmutablePlatform(this.groupId, this.groupIdJson, this.artifactId, newValue, this.releases);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Platform#getReleases() releases}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutablePlatform withReleases(Release... elements) {
    List<Release> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return new ImmutablePlatform(this.groupId, this.groupIdJson, this.artifactId, this.artifactIdJson, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Platform#getReleases() releases}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of releases elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutablePlatform withReleases(Iterable<? extends Release> elements) {
    if (this.releases == elements) return this;
    List<Release> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return new ImmutablePlatform(this.groupId, this.groupIdJson, this.artifactId, this.artifactIdJson, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutablePlatform} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutablePlatform
        && equalTo((ImmutablePlatform) another);
  }

  private boolean equalTo(ImmutablePlatform another) {
    return groupId.equals(another.groupId)
        && artifactId.equals(another.artifactId);
  }

  /**
   * Computes a hash code from attributes: {@code groupId}, {@code artifactId}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + groupId.hashCode();
    h += (h << 5) + artifactId.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Platform} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Platform{"
        + "groupId=" + groupId
        + ", artifactId=" + artifactId
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Platform {
    String groupId;
    String groupIdJson;
    String artifactId;
    String artifactIdJson;
    List<Release> releases = Collections.emptyList();
    @JsonProperty("group-id")
    public void setGroupId(String groupId) {
      this.groupId = groupId;
    }
    @JsonProperty("group-id-json")
    public void setGroupIdJson(String groupIdJson) {
      this.groupIdJson = groupIdJson;
    }
    @JsonProperty("artifact-id")
    public void setArtifactId(String artifactId) {
      this.artifactId = artifactId;
    }
    @JsonProperty("artifact-id-json")
    public void setArtifactIdJson(String artifactIdJson) {
      this.artifactIdJson = artifactIdJson;
    }
    @JsonProperty("releases")
    public void setReleases(List<Release> releases) {
      this.releases = releases;
    }
    @Override
    public String getGroupId() { throw new UnsupportedOperationException(); }
    @Override
    public String getGroupIdJson() { throw new UnsupportedOperationException(); }
    @Override
    public String getArtifactId() { throw new UnsupportedOperationException(); }
    @Override
    public String getArtifactIdJson() { throw new UnsupportedOperationException(); }
    @Override
    public List<Release> getReleases() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutablePlatform fromJson(Json json) {
    ImmutablePlatform.Builder builder = ImmutablePlatform.builder();
    if (json.groupId != null) {
      builder.groupId(json.groupId);
    }
    if (json.groupIdJson != null) {
      builder.groupIdJson(json.groupIdJson);
    }
    if (json.artifactId != null) {
      builder.artifactId(json.artifactId);
    }
    if (json.artifactIdJson != null) {
      builder.artifactIdJson(json.artifactIdJson);
    }
    if (json.releases != null) {
      builder.addAllReleases(json.releases);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Platform} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Platform instance
   */
  public static ImmutablePlatform copyOf(Platform instance) {
    if (instance instanceof ImmutablePlatform) {
      return (ImmutablePlatform) instance;
    }
    return ImmutablePlatform.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutablePlatform ImmutablePlatform}.
   * <pre>
   * ImmutablePlatform.builder()
   *    .groupId(String) // required {@link Platform#getGroupId() groupId}
   *    .groupIdJson(String) // optional {@link Platform#getGroupIdJson() groupIdJson}
   *    .artifactId(String) // required {@link Platform#getArtifactId() artifactId}
   *    .artifactIdJson(String) // optional {@link Platform#getArtifactIdJson() artifactIdJson}
   *    .addReleases|addAllReleases(io.quarkus.registry.model.Release) // {@link Platform#getReleases() releases} elements
   *    .build();
   * </pre>
   * @return A new ImmutablePlatform builder
   */
  public static ImmutablePlatform.Builder builder() {
    return new ImmutablePlatform.Builder();
  }

  /**
   * Builds instances of type {@link ImmutablePlatform ImmutablePlatform}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_GROUP_ID = 0x1L;
    private static final long INIT_BIT_ARTIFACT_ID = 0x2L;
    private long initBits = 0x3L;

    private String groupId;
    private String groupIdJson;
    private String artifactId;
    private String artifactIdJson;
    private List<Release> releases = new ArrayList<Release>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Platform} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Platform instance) {
      Objects.requireNonNull(instance, "instance");
      groupId(instance.getGroupId());
      groupIdJson(instance.getGroupIdJson());
      artifactId(instance.getArtifactId());
      artifactIdJson(instance.getArtifactIdJson());
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Initializes the value for the {@link Platform#getGroupId() groupId} attribute.
     * @param groupId The value for groupId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("group-id")
    public final Builder groupId(String groupId) {
      this.groupId = Objects.requireNonNull(groupId, "groupId");
      initBits &= ~INIT_BIT_GROUP_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link Platform#getGroupIdJson() groupIdJson} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link Platform#getGroupIdJson() groupIdJson}.</em>
     * @param groupIdJson The value for groupIdJson 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("group-id-json")
    public final Builder groupIdJson(String groupIdJson) {
      this.groupIdJson = Objects.requireNonNull(groupIdJson, "groupIdJson");
      return this;
    }

    /**
     * Initializes the value for the {@link Platform#getArtifactId() artifactId} attribute.
     * @param artifactId The value for artifactId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("artifact-id")
    public final Builder artifactId(String artifactId) {
      this.artifactId = Objects.requireNonNull(artifactId, "artifactId");
      initBits &= ~INIT_BIT_ARTIFACT_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link Platform#getArtifactIdJson() artifactIdJson} attribute.
     * <p><em>If not set, this attribute will have a default value as returned by the initializer of {@link Platform#getArtifactIdJson() artifactIdJson}.</em>
     * @param artifactIdJson The value for artifactIdJson 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("artifact-id-json")
    public final Builder artifactIdJson(String artifactIdJson) {
      this.artifactIdJson = Objects.requireNonNull(artifactIdJson, "artifactIdJson");
      return this;
    }

    /**
     * Adds one element to {@link Platform#getReleases() releases} list.
     * @param element A releases element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release element) {
      this.releases.add(Objects.requireNonNull(element, "releases element"));
      return this;
    }

    /**
     * Adds elements to {@link Platform#getReleases() releases} list.
     * @param elements An array of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release... elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Platform#getReleases() releases} list.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("releases")
    public final Builder releases(Iterable<? extends Release> elements) {
      this.releases.clear();
      return addAllReleases(elements);
    }

    /**
     * Adds elements to {@link Platform#getReleases() releases} list.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllReleases(Iterable<? extends Release> elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutablePlatform ImmutablePlatform}.
     * @return An immutable instance of Platform
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutablePlatform build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutablePlatform(this);
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_GROUP_ID) != 0) attributes.add("groupId");
      if ((initBits & INIT_BIT_ARTIFACT_ID) != 0) attributes.add("artifactId");
      return "Cannot build Platform, some of required attributes are not set " + attributes;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
