package io.quarkus.registry.catalog.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link Repository}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableRepository.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableRepository extends Repository {
  private final List<Extension> individualExtensions;
  private final List<Platform> platforms;

  private ImmutableRepository(
      List<Extension> individualExtensions,
      List<Platform> platforms) {
    this.individualExtensions = individualExtensions;
    this.platforms = platforms;
  }

  /**
   * @return The value of the {@code individualExtensions} attribute
   */
  @JsonProperty("individualExtensions")
  @Override
  public List<Extension> getIndividualExtensions() {
    return individualExtensions;
  }

  /**
   * @return The value of the {@code platforms} attribute
   */
  @JsonProperty("platforms")
  @Override
  public List<Platform> getPlatforms() {
    return platforms;
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Repository#getIndividualExtensions() individualExtensions}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRepository withIndividualExtensions(Extension... elements) {
    List<Extension> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableRepository(newValue, this.platforms);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Repository#getIndividualExtensions() individualExtensions}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of individualExtensions elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRepository withIndividualExtensions(Iterable<? extends Extension> elements) {
    if (this.individualExtensions == elements) return this;
    List<Extension> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return new ImmutableRepository(newValue, this.platforms);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Repository#getPlatforms() platforms}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRepository withPlatforms(Platform... elements) {
    List<Platform> newValue = createUnmodifiableList(false, createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableRepository(this.individualExtensions, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Repository#getPlatforms() platforms}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of platforms elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRepository withPlatforms(Iterable<? extends Platform> elements) {
    if (this.platforms == elements) return this;
    List<Platform> newValue = createUnmodifiableList(false, createSafeList(elements, true, false));
    return new ImmutableRepository(this.individualExtensions, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableRepository} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableRepository
        && equalTo((ImmutableRepository) another);
  }

  private boolean equalTo(ImmutableRepository another) {
    return individualExtensions.equals(another.individualExtensions)
        && platforms.equals(another.platforms);
  }

  /**
   * Computes a hash code from attributes: {@code individualExtensions}, {@code platforms}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + individualExtensions.hashCode();
    h += (h << 5) + platforms.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Repository} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Repository{"
        + "individualExtensions=" + individualExtensions
        + ", platforms=" + platforms
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json extends Repository {
    List<Extension> individualExtensions = Collections.emptyList();
    List<Platform> platforms = Collections.emptyList();
    @JsonProperty("individualExtensions")
    public void setIndividualExtensions(List<Extension> individualExtensions) {
      this.individualExtensions = individualExtensions;
    }
    @JsonProperty("platforms")
    public void setPlatforms(List<Platform> platforms) {
      this.platforms = platforms;
    }
    @Override
    public List<Extension> getIndividualExtensions() { throw new UnsupportedOperationException(); }
    @Override
    public List<Platform> getPlatforms() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableRepository fromJson(Json json) {
    ImmutableRepository.Builder builder = ImmutableRepository.builder();
    if (json.individualExtensions != null) {
      builder.addAllIndividualExtensions(json.individualExtensions);
    }
    if (json.platforms != null) {
      builder.addAllPlatforms(json.platforms);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Repository} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Repository instance
   */
  public static ImmutableRepository copyOf(Repository instance) {
    if (instance instanceof ImmutableRepository) {
      return (ImmutableRepository) instance;
    }
    return ImmutableRepository.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableRepository ImmutableRepository}.
   * <pre>
   * ImmutableRepository.builder()
   *    .addIndividualExtensions|addAllIndividualExtensions(io.quarkus.registry.catalog.model.Extension) // {@link Repository#getIndividualExtensions() individualExtensions} elements
   *    .addPlatforms|addAllPlatforms(io.quarkus.registry.catalog.model.Platform) // {@link Repository#getPlatforms() platforms} elements
   *    .build();
   * </pre>
   * @return A new ImmutableRepository builder
   */
  public static ImmutableRepository.Builder builder() {
    return new ImmutableRepository.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableRepository ImmutableRepository}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private List<Extension> individualExtensions = new ArrayList<Extension>();
    private List<Platform> platforms = new ArrayList<Platform>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Repository} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Repository instance) {
      Objects.requireNonNull(instance, "instance");
      addAllIndividualExtensions(instance.getIndividualExtensions());
      addAllPlatforms(instance.getPlatforms());
      return this;
    }

    /**
     * Adds one element to {@link Repository#getIndividualExtensions() individualExtensions} list.
     * @param element A individualExtensions element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addIndividualExtensions(Extension element) {
      this.individualExtensions.add(Objects.requireNonNull(element, "individualExtensions element"));
      return this;
    }

    /**
     * Adds elements to {@link Repository#getIndividualExtensions() individualExtensions} list.
     * @param elements An array of individualExtensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addIndividualExtensions(Extension... elements) {
      for (Extension element : elements) {
        this.individualExtensions.add(Objects.requireNonNull(element, "individualExtensions element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Repository#getIndividualExtensions() individualExtensions} list.
     * @param elements An iterable of individualExtensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("individualExtensions")
    public final Builder individualExtensions(Iterable<? extends Extension> elements) {
      this.individualExtensions.clear();
      return addAllIndividualExtensions(elements);
    }

    /**
     * Adds elements to {@link Repository#getIndividualExtensions() individualExtensions} list.
     * @param elements An iterable of individualExtensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllIndividualExtensions(Iterable<? extends Extension> elements) {
      for (Extension element : elements) {
        this.individualExtensions.add(Objects.requireNonNull(element, "individualExtensions element"));
      }
      return this;
    }

    /**
     * Adds one element to {@link Repository#getPlatforms() platforms} list.
     * @param element A platforms element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Platform element) {
      this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      return this;
    }

    /**
     * Adds elements to {@link Repository#getPlatforms() platforms} list.
     * @param elements An array of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Platform... elements) {
      for (Platform element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Repository#getPlatforms() platforms} list.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("platforms")
    public final Builder platforms(Iterable<? extends Platform> elements) {
      this.platforms.clear();
      return addAllPlatforms(elements);
    }

    /**
     * Adds elements to {@link Repository#getPlatforms() platforms} list.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllPlatforms(Iterable<? extends Platform> elements) {
      for (Platform element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutableRepository ImmutableRepository}.
     * @return An immutable instance of Repository
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableRepository build() {
      return new ImmutableRepository(createUnmodifiableList(true, individualExtensions), createUnmodifiableList(true, platforms));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T> List<T> createUnmodifiableList(boolean clone, List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptyList();
    case 1: return Collections.singletonList(list.get(0));
    default:
      if (clone) {
        return Collections.unmodifiableList(new ArrayList<>(list));
      } else {
        if (list instanceof ArrayList<?>) {
          ((ArrayList<?>) list).trimToSize();
        }
        return Collections.unmodifiableList(list);
      }
    }
  }
}
