package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link ArtifactCoords}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableArtifactCoords.builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableArtifactCoords.of()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableArtifactCoords implements ArtifactCoords {
  private final ArtifactKey id;
  private final String version;

  private ImmutableArtifactCoords(ArtifactKey id, String version) {
    this.id = Objects.requireNonNull(id, "id");
    this.version = Objects.requireNonNull(version, "version");
  }

  private ImmutableArtifactCoords(
      ImmutableArtifactCoords original,
      ArtifactKey id,
      String version) {
    this.id = id;
    this.version = version;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @JsonUnwrapped
  @Override
  public ArtifactKey getId() {
    return id;
  }

  /**
   * @return The value of the {@code version} attribute
   */
  @JsonProperty("version")
  @Override
  public String getVersion() {
    return version;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ArtifactCoords#getId() id} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for id
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableArtifactCoords withId(ArtifactKey value) {
    if (this.id == value) return this;
    ArtifactKey newValue = Objects.requireNonNull(value, "id");
    return new ImmutableArtifactCoords(this, newValue, this.version);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ArtifactCoords#getVersion() version} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for version
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableArtifactCoords withVersion(String value) {
    String newValue = Objects.requireNonNull(value, "version");
    if (this.version.equals(newValue)) return this;
    return new ImmutableArtifactCoords(this, this.id, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableArtifactCoords} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableArtifactCoords
        && equalTo((ImmutableArtifactCoords) another);
  }

  private boolean equalTo(ImmutableArtifactCoords another) {
    return id.equals(another.id)
        && version.equals(another.version);
  }

  /**
   * Computes a hash code from attributes: {@code id}, {@code version}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + id.hashCode();
    h += (h << 5) + version.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ArtifactCoords} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "ArtifactCoords{"
        + "id=" + id
        + ", version=" + version
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements ArtifactCoords {
    ArtifactKey id;
    String version;
    @JsonProperty("id")
    @JsonUnwrapped
    public void setId(ArtifactKey id) {
      this.id = id;
    }
    @JsonProperty("version")
    public void setVersion(String version) {
      this.version = version;
    }
    @Override
    public ArtifactKey getId() { throw new UnsupportedOperationException(); }
    @Override
    public String getVersion() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableArtifactCoords fromJson(Json json) {
    ImmutableArtifactCoords.Builder builder = ImmutableArtifactCoords.builder();
    if (json.id != null) {
      builder.id(json.id);
    }
    if (json.version != null) {
      builder.version(json.version);
    }
    return builder.build();
  }

  /**
   * Construct a new immutable {@code ArtifactCoords} instance.
   * @param id The value for the {@code id} attribute
   * @param version The value for the {@code version} attribute
   * @return An immutable ArtifactCoords instance
   */
  public static ImmutableArtifactCoords of(ArtifactKey id, String version) {
    return new ImmutableArtifactCoords(id, version);
  }

  /**
   * Creates an immutable copy of a {@link ArtifactCoords} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ArtifactCoords instance
   */
  public static ImmutableArtifactCoords copyOf(ArtifactCoords instance) {
    if (instance instanceof ImmutableArtifactCoords) {
      return (ImmutableArtifactCoords) instance;
    }
    return ImmutableArtifactCoords.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableArtifactCoords ImmutableArtifactCoords}.
   * <pre>
   * ImmutableArtifactCoords.builder()
   *    .id(io.quarkus.registry.model.ArtifactKey) // required {@link ArtifactCoords#getId() id}
   *    .version(String) // required {@link ArtifactCoords#getVersion() version}
   *    .build();
   * </pre>
   * @return A new ImmutableArtifactCoords builder
   */
  public static ImmutableArtifactCoords.Builder builder() {
    return new ImmutableArtifactCoords.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableArtifactCoords ImmutableArtifactCoords}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_ID = 0x1L;
    private static final long INIT_BIT_VERSION = 0x2L;
    private long initBits = 0x3L;

    private ArtifactKey id;
    private String version;

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ArtifactCoords} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(ArtifactCoords instance) {
      Objects.requireNonNull(instance, "instance");
      id(instance.getId());
      version(instance.getVersion());
      return this;
    }

    /**
     * Initializes the value for the {@link ArtifactCoords#getId() id} attribute.
     * @param id The value for id 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    @JsonUnwrapped
    public final Builder id(ArtifactKey id) {
      this.id = Objects.requireNonNull(id, "id");
      initBits &= ~INIT_BIT_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link ArtifactCoords#getVersion() version} attribute.
     * @param version The value for version 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("version")
    public final Builder version(String version) {
      this.version = Objects.requireNonNull(version, "version");
      initBits &= ~INIT_BIT_VERSION;
      return this;
    }

    /**
     * Builds a new {@link ImmutableArtifactCoords ImmutableArtifactCoords}.
     * @return An immutable instance of ArtifactCoords
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableArtifactCoords build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableArtifactCoords(null, id, version);
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_ID) != 0) attributes.add("id");
      if ((initBits & INIT_BIT_VERSION) != 0) attributes.add("version");
      return "Cannot build ArtifactCoords, some of required attributes are not set " + attributes;
    }
  }
}
