package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link ArtifactKey}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableArtifactKey.builder()}.
 * Use the static factory method to create immutable instances:
 * {@code ImmutableArtifactKey.of()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableArtifactKey implements ArtifactKey {
  private final String groupId;
  private final String artifactId;

  private ImmutableArtifactKey(String groupId, String artifactId) {
    this.groupId = Objects.requireNonNull(groupId, "groupId");
    this.artifactId = Objects.requireNonNull(artifactId, "artifactId");
  }

  private ImmutableArtifactKey(ImmutableArtifactKey original, String groupId, String artifactId) {
    this.groupId = groupId;
    this.artifactId = artifactId;
  }

  /**
   * @return The value of the {@code groupId} attribute
   */
  @JsonProperty("group-id")
  @Override
  public String getGroupId() {
    return groupId;
  }

  /**
   * @return The value of the {@code artifactId} attribute
   */
  @JsonProperty("artifact-id")
  @Override
  public String getArtifactId() {
    return artifactId;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ArtifactKey#getGroupId() groupId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for groupId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableArtifactKey withGroupId(String value) {
    String newValue = Objects.requireNonNull(value, "groupId");
    if (this.groupId.equals(newValue)) return this;
    return new ImmutableArtifactKey(this, newValue, this.artifactId);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link ArtifactKey#getArtifactId() artifactId} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for artifactId
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableArtifactKey withArtifactId(String value) {
    String newValue = Objects.requireNonNull(value, "artifactId");
    if (this.artifactId.equals(newValue)) return this;
    return new ImmutableArtifactKey(this, this.groupId, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableArtifactKey} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableArtifactKey
        && equalTo((ImmutableArtifactKey) another);
  }

  private boolean equalTo(ImmutableArtifactKey another) {
    return groupId.equals(another.groupId)
        && artifactId.equals(another.artifactId);
  }

  /**
   * Computes a hash code from attributes: {@code groupId}, {@code artifactId}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + groupId.hashCode();
    h += (h << 5) + artifactId.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ArtifactKey} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "ArtifactKey{"
        + "groupId=" + groupId
        + ", artifactId=" + artifactId
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements ArtifactKey {
    String groupId;
    String artifactId;
    @JsonProperty("group-id")
    public void setGroupId(String groupId) {
      this.groupId = groupId;
    }
    @JsonProperty("artifact-id")
    public void setArtifactId(String artifactId) {
      this.artifactId = artifactId;
    }
    @Override
    public String getGroupId() { throw new UnsupportedOperationException(); }
    @Override
    public String getArtifactId() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableArtifactKey fromJson(Json json) {
    ImmutableArtifactKey.Builder builder = ImmutableArtifactKey.builder();
    if (json.groupId != null) {
      builder.groupId(json.groupId);
    }
    if (json.artifactId != null) {
      builder.artifactId(json.artifactId);
    }
    return builder.build();
  }

  /**
   * Construct a new immutable {@code ArtifactKey} instance.
   * @param groupId The value for the {@code groupId} attribute
   * @param artifactId The value for the {@code artifactId} attribute
   * @return An immutable ArtifactKey instance
   */
  public static ImmutableArtifactKey of(String groupId, String artifactId) {
    return new ImmutableArtifactKey(groupId, artifactId);
  }

  /**
   * Creates an immutable copy of a {@link ArtifactKey} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ArtifactKey instance
   */
  public static ImmutableArtifactKey copyOf(ArtifactKey instance) {
    if (instance instanceof ImmutableArtifactKey) {
      return (ImmutableArtifactKey) instance;
    }
    return ImmutableArtifactKey.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableArtifactKey ImmutableArtifactKey}.
   * <pre>
   * ImmutableArtifactKey.builder()
   *    .groupId(String) // required {@link ArtifactKey#getGroupId() groupId}
   *    .artifactId(String) // required {@link ArtifactKey#getArtifactId() artifactId}
   *    .build();
   * </pre>
   * @return A new ImmutableArtifactKey builder
   */
  public static ImmutableArtifactKey.Builder builder() {
    return new ImmutableArtifactKey.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableArtifactKey ImmutableArtifactKey}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_GROUP_ID = 0x1L;
    private static final long INIT_BIT_ARTIFACT_ID = 0x2L;
    private long initBits = 0x3L;

    private String groupId;
    private String artifactId;

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ArtifactKey} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(ArtifactKey instance) {
      Objects.requireNonNull(instance, "instance");
      groupId(instance.getGroupId());
      artifactId(instance.getArtifactId());
      return this;
    }

    /**
     * Initializes the value for the {@link ArtifactKey#getGroupId() groupId} attribute.
     * @param groupId The value for groupId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("group-id")
    public final Builder groupId(String groupId) {
      this.groupId = Objects.requireNonNull(groupId, "groupId");
      initBits &= ~INIT_BIT_GROUP_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link ArtifactKey#getArtifactId() artifactId} attribute.
     * @param artifactId The value for artifactId 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("artifact-id")
    public final Builder artifactId(String artifactId) {
      this.artifactId = Objects.requireNonNull(artifactId, "artifactId");
      initBits &= ~INIT_BIT_ARTIFACT_ID;
      return this;
    }

    /**
     * Builds a new {@link ImmutableArtifactKey ImmutableArtifactKey}.
     * @return An immutable instance of ArtifactKey
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableArtifactKey build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableArtifactKey(null, groupId, artifactId);
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_GROUP_ID) != 0) attributes.add("groupId");
      if ((initBits & INIT_BIT_ARTIFACT_ID) != 0) attributes.add("artifactId");
      return "Cannot build ArtifactKey, some of required attributes are not set " + attributes;
    }
  }
}
