package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Objects;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Immutable implementation of {@link Extension}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableExtension.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableExtension implements Extension {
  private final ArtifactKey id;
  private final String name;
  private final @Nullable String description;
  private final Map<String, Object> metadata;
  private final SortedSet<Extension.ExtensionRelease> releases;

  private ImmutableExtension(
      ArtifactKey id,
      String name,
      @Nullable String description,
      Map<String, Object> metadata,
      SortedSet<Extension.ExtensionRelease> releases) {
    this.id = id;
    this.name = name;
    this.description = description;
    this.metadata = metadata;
    this.releases = releases;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @JsonUnwrapped
  @Override
  public ArtifactKey getId() {
    return id;
  }

  /**
   * @return The value of the {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public String getName() {
    return name;
  }

  /**
   * @return The value of the {@code description} attribute
   */
  @JsonProperty("description")
  @Override
  public @Nullable String getDescription() {
    return description;
  }

  /**
   * @return The value of the {@code metadata} attribute
   */
  @JsonProperty("metadata")
  @Override
  public Map<String, Object> getMetadata() {
    return metadata;
  }

  /**
   * @return The value of the {@code releases} attribute
   */
  @JsonProperty("releases")
  @Override
  public SortedSet<Extension.ExtensionRelease> getReleases() {
    return releases;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getId() id} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for id
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withId(ArtifactKey value) {
    if (this.id == value) return this;
    ArtifactKey newValue = Objects.requireNonNull(value, "id");
    return new ImmutableExtension(newValue, this.name, this.description, this.metadata, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getName() name} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for name
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withName(String value) {
    String newValue = Objects.requireNonNull(value, "name");
    if (this.name.equals(newValue)) return this;
    return new ImmutableExtension(this.id, newValue, this.description, this.metadata, this.releases);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension#getDescription() description} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for description (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtension withDescription(@Nullable String value) {
    if (Objects.equals(this.description, value)) return this;
    return new ImmutableExtension(this.id, this.name, value, this.metadata, this.releases);
  }

  /**
   * Copy the current immutable object by replacing the {@link Extension#getMetadata() metadata} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the metadata map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withMetadata(Map<String, ? extends Object> entries) {
    if (this.metadata == entries) return this;
    Map<String, Object> newValue = createUnmodifiableMap(true, false, entries);
    return new ImmutableExtension(this.id, this.name, this.description, newValue, this.releases);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getReleases() releases}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withReleases(Extension.ExtensionRelease... elements) {
    SortedSet<Extension.ExtensionRelease> newValue = createUnmodifiableSortedSet(true, createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableExtension(this.id, this.name, this.description, this.metadata, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension#getReleases() releases}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of releases elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtension withReleases(Iterable<? extends Extension.ExtensionRelease> elements) {
    if (this.releases == elements) return this;
    SortedSet<Extension.ExtensionRelease> newValue = createUnmodifiableSortedSet(true, createSafeList(elements, true, false));
    return new ImmutableExtension(this.id, this.name, this.description, this.metadata, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtension} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtension
        && equalTo((ImmutableExtension) another);
  }

  private boolean equalTo(ImmutableExtension another) {
    return id.equals(another.id);
  }

  /**
   * Computes a hash code from attributes: {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + id.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Extension} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Extension{"
        + "id=" + id
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Extension {
    ArtifactKey id;
    String name;
    String description;
    Map<String, Object> metadata = Collections.emptyMap();
    SortedSet<Extension.ExtensionRelease> releases = createUnmodifiableSortedSet(true, Collections.<Extension.ExtensionRelease>emptyList());
    @JsonProperty("id")
    @JsonUnwrapped
    public void setId(ArtifactKey id) {
      this.id = id;
    }
    @JsonProperty("name")
    public void setName(String name) {
      this.name = name;
    }
    @JsonProperty("description")
    public void setDescription(@Nullable String description) {
      this.description = description;
    }
    @JsonProperty("metadata")
    public void setMetadata(Map<String, Object> metadata) {
      this.metadata = metadata;
    }
    @JsonProperty("releases")
    public void setReleases(SortedSet<Extension.ExtensionRelease> releases) {
      this.releases = releases;
    }
    @Override
    public ArtifactKey getId() { throw new UnsupportedOperationException(); }
    @Override
    public String getName() { throw new UnsupportedOperationException(); }
    @Override
    public String getDescription() { throw new UnsupportedOperationException(); }
    @Override
    public Map<String, Object> getMetadata() { throw new UnsupportedOperationException(); }
    @Override
    public SortedSet<Extension.ExtensionRelease> getReleases() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableExtension fromJson(Json json) {
    ImmutableExtension.Builder builder = ImmutableExtension.builder();
    if (json.id != null) {
      builder.id(json.id);
    }
    if (json.name != null) {
      builder.name(json.name);
    }
    if (json.description != null) {
      builder.description(json.description);
    }
    if (json.metadata != null) {
      builder.putAllMetadata(json.metadata);
    }
    if (json.releases != null) {
      builder.addAllReleases(json.releases);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Extension} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Extension instance
   */
  public static ImmutableExtension copyOf(Extension instance) {
    if (instance instanceof ImmutableExtension) {
      return (ImmutableExtension) instance;
    }
    return ImmutableExtension.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableExtension ImmutableExtension}.
   * <pre>
   * ImmutableExtension.builder()
   *    .id(io.quarkus.registry.model.ArtifactKey) // required {@link Extension#getId() id}
   *    .name(String) // required {@link Extension#getName() name}
   *    .description(String | null) // nullable {@link Extension#getDescription() description}
   *    .putMetadata|putAllMetadata(String =&gt; Object) // {@link Extension#getMetadata() metadata} mappings
   *    .addReleases|addAllReleases(io.quarkus.registry.model.Extension.ExtensionRelease) // {@link Extension#getReleases() releases} elements
   *    .build();
   * </pre>
   * @return A new ImmutableExtension builder
   */
  public static ImmutableExtension.Builder builder() {
    return new ImmutableExtension.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableExtension ImmutableExtension}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_ID = 0x1L;
    private static final long INIT_BIT_NAME = 0x2L;
    private long initBits = 0x3L;

    private ArtifactKey id;
    private String name;
    private String description;
    private Map<String, Object> metadata = new LinkedHashMap<String, Object>();
    private List<Extension.ExtensionRelease> releases = new ArrayList<Extension.ExtensionRelease>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ModifiableExtension} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(ModifiableExtension instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance.idIsSet()) {
        id(instance.getId());
      }
      if (instance.nameIsSet()) {
        name(instance.getName());
      }
      @Nullable String descriptionValue = instance.getDescription();
      if (descriptionValue != null) {
        description(descriptionValue);
      }
      putAllMetadata(instance.getMetadata());
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code Extension} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Extension instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance instanceof ModifiableExtension) {
        return from((ModifiableExtension) instance);
      }
      id(instance.getId());
      name(instance.getName());
      @Nullable String descriptionValue = instance.getDescription();
      if (descriptionValue != null) {
        description(descriptionValue);
      }
      putAllMetadata(instance.getMetadata());
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Initializes the value for the {@link Extension#getId() id} attribute.
     * @param id The value for id 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    @JsonUnwrapped
    public final Builder id(ArtifactKey id) {
      this.id = Objects.requireNonNull(id, "id");
      initBits &= ~INIT_BIT_ID;
      return this;
    }

    /**
     * Initializes the value for the {@link Extension#getName() name} attribute.
     * @param name The value for name 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("name")
    public final Builder name(String name) {
      this.name = Objects.requireNonNull(name, "name");
      initBits &= ~INIT_BIT_NAME;
      return this;
    }

    /**
     * Initializes the value for the {@link Extension#getDescription() description} attribute.
     * @param description The value for description (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("description")
    public final Builder description(@Nullable String description) {
      this.description = description;
      return this;
    }

    /**
     * Put one entry to the {@link Extension#getMetadata() metadata} map.
     * @param key The key in the metadata map
     * @param value The associated value in the metadata map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putMetadata(String key, Object value) {
      this.metadata.put(
          Objects.requireNonNull(key, "metadata key"),
          Objects.requireNonNull(value, "metadata value"));
      return this;
    }

    /**
     * Put one entry to the {@link Extension#getMetadata() metadata} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putMetadata(Map.Entry<String, ? extends Object> entry) {
      String k = entry.getKey();
      Object v = entry.getValue();
      this.metadata.put(
          Objects.requireNonNull(k, "metadata key"),
          Objects.requireNonNull(v, "metadata value"));
      return this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link Extension#getMetadata() metadata} map. Nulls are not permitted
     * @param entries The entries that will be added to the metadata map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("metadata")
    public final Builder metadata(Map<String, ? extends Object> entries) {
      this.metadata.clear();
      return putAllMetadata(entries);
    }

    /**
     * Put all mappings from the specified map as entries to {@link Extension#getMetadata() metadata} map. Nulls are not permitted
     * @param entries The entries that will be added to the metadata map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putAllMetadata(Map<String, ? extends Object> entries) {
      for (Map.Entry<String, ? extends Object> e : entries.entrySet()) {
        String k = e.getKey();
        Object v = e.getValue();
        this.metadata.put(
            Objects.requireNonNull(k, "metadata key"),
            Objects.requireNonNull(v, "metadata value"));
      }
      return this;
    }

    /**
     * Adds one element to {@link Extension#getReleases() releases} sortedSet.
     * @param element A releases element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Extension.ExtensionRelease element) {
      this.releases.add(Objects.requireNonNull(element, "releases element"));
      return this;
    }

    /**
     * Adds elements to {@link Extension#getReleases() releases} sortedSet.
     * @param elements An array of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Extension.ExtensionRelease... elements) {
      for (Extension.ExtensionRelease element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Extension#getReleases() releases} sortedSet.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("releases")
    public final Builder releases(Iterable<? extends Extension.ExtensionRelease> elements) {
      this.releases.clear();
      return addAllReleases(elements);
    }

    /**
     * Adds elements to {@link Extension#getReleases() releases} sortedSet.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllReleases(Iterable<? extends Extension.ExtensionRelease> elements) {
      for (Extension.ExtensionRelease element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutableExtension ImmutableExtension}.
     * @return An immutable instance of Extension
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableExtension build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableExtension(
          id,
          name,
          description,
          createUnmodifiableMap(false, false, metadata),
          createUnmodifiableSortedSet(true, createSafeList(releases, false, false)));
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_ID) != 0) attributes.add("id");
      if ((initBits & INIT_BIT_NAME) != 0) attributes.add("name");
      return "Cannot build Extension, some of required attributes are not set " + attributes;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T extends Comparable<T>> NavigableSet<T> createUnmodifiableSortedSet(boolean reverse, List<T> list) {
    TreeSet<T> set = reverse
        ? new TreeSet<>(Collections.reverseOrder())
        : new TreeSet<>();
    set.addAll(list);
    return Collections.unmodifiableNavigableSet(set);
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
