package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * Immutable implementation of {@link Extension.ExtensionRelease}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableExtensionRelease.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableExtensionRelease implements Extension.ExtensionRelease {
  private final Release release;
  private final Set<Extension.ExtensionPlatformRelease> platforms;

  private ImmutableExtensionRelease(
      Release release,
      Set<Extension.ExtensionPlatformRelease> platforms) {
    this.release = release;
    this.platforms = platforms;
  }

  /**
   * @return The value of the {@code release} attribute
   */
  @JsonProperty("release")
  @JsonUnwrapped
  @Override
  public Release getRelease() {
    return release;
  }

  /**
   * @return The value of the {@code platforms} attribute
   */
  @JsonProperty("platforms")
  @Override
  public Set<Extension.ExtensionPlatformRelease> getPlatforms() {
    return platforms;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Extension.ExtensionRelease#getRelease() release} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for release
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableExtensionRelease withRelease(Release value) {
    if (this.release == value) return this;
    Release newValue = Objects.requireNonNull(value, "release");
    return new ImmutableExtensionRelease(newValue, this.platforms);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension.ExtensionRelease#getPlatforms() platforms}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionRelease withPlatforms(Extension.ExtensionPlatformRelease... elements) {
    Set<Extension.ExtensionPlatformRelease> newValue = createUnmodifiableSet(createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableExtensionRelease(this.release, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Extension.ExtensionRelease#getPlatforms() platforms}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of platforms elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableExtensionRelease withPlatforms(Iterable<? extends Extension.ExtensionPlatformRelease> elements) {
    if (this.platforms == elements) return this;
    Set<Extension.ExtensionPlatformRelease> newValue = createUnmodifiableSet(createSafeList(elements, true, false));
    return new ImmutableExtensionRelease(this.release, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableExtensionRelease} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableExtensionRelease
        && equalTo((ImmutableExtensionRelease) another);
  }

  private boolean equalTo(ImmutableExtensionRelease another) {
    return release.equals(another.release);
  }

  /**
   * Computes a hash code from attributes: {@code release}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + release.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code ExtensionRelease} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "ExtensionRelease{"
        + "release=" + release
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Extension.ExtensionRelease {
    Release release;
    Set<Extension.ExtensionPlatformRelease> platforms = Collections.emptySet();
    @JsonProperty("release")
    @JsonUnwrapped
    public void setRelease(Release release) {
      this.release = release;
    }
    @JsonProperty("platforms")
    public void setPlatforms(Set<Extension.ExtensionPlatformRelease> platforms) {
      this.platforms = platforms;
    }
    @Override
    public Release getRelease() { throw new UnsupportedOperationException(); }
    @Override
    public Set<Extension.ExtensionPlatformRelease> getPlatforms() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableExtensionRelease fromJson(Json json) {
    ImmutableExtensionRelease.Builder builder = ImmutableExtensionRelease.builder();
    if (json.release != null) {
      builder.release(json.release);
    }
    if (json.platforms != null) {
      builder.addAllPlatforms(json.platforms);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Extension.ExtensionRelease} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable ExtensionRelease instance
   */
  public static ImmutableExtensionRelease copyOf(Extension.ExtensionRelease instance) {
    if (instance instanceof ImmutableExtensionRelease) {
      return (ImmutableExtensionRelease) instance;
    }
    return ImmutableExtensionRelease.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableExtensionRelease ImmutableExtensionRelease}.
   * <pre>
   * ImmutableExtensionRelease.builder()
   *    .release(io.quarkus.registry.model.Release) // required {@link Extension.ExtensionRelease#getRelease() release}
   *    .addPlatforms|addAllPlatforms(io.quarkus.registry.model.Extension.ExtensionPlatformRelease) // {@link Extension.ExtensionRelease#getPlatforms() platforms} elements
   *    .build();
   * </pre>
   * @return A new ImmutableExtensionRelease builder
   */
  public static ImmutableExtensionRelease.Builder builder() {
    return new ImmutableExtensionRelease.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableExtensionRelease ImmutableExtensionRelease}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_RELEASE = 0x1L;
    private long initBits = 0x1L;

    private Release release;
    private List<Extension.ExtensionPlatformRelease> platforms = new ArrayList<Extension.ExtensionPlatformRelease>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ModifiableExtensionRelease} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(ModifiableExtensionRelease instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance.releaseIsSet()) {
        release(instance.getRelease());
      }
      addAllPlatforms(instance.getPlatforms());
      return this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code ExtensionRelease} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Extension.ExtensionRelease instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance instanceof ModifiableExtensionRelease) {
        return from((ModifiableExtensionRelease) instance);
      }
      release(instance.getRelease());
      addAllPlatforms(instance.getPlatforms());
      return this;
    }

    /**
     * Initializes the value for the {@link Extension.ExtensionRelease#getRelease() release} attribute.
     * @param release The value for release 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("release")
    @JsonUnwrapped
    public final Builder release(Release release) {
      this.release = Objects.requireNonNull(release, "release");
      initBits &= ~INIT_BIT_RELEASE;
      return this;
    }

    /**
     * Adds one element to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
     * @param element A platforms element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Extension.ExtensionPlatformRelease element) {
      this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      return this;
    }

    /**
     * Adds elements to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
     * @param elements An array of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Extension.ExtensionPlatformRelease... elements) {
      for (Extension.ExtensionPlatformRelease element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("platforms")
    public final Builder platforms(Iterable<? extends Extension.ExtensionPlatformRelease> elements) {
      this.platforms.clear();
      return addAllPlatforms(elements);
    }

    /**
     * Adds elements to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllPlatforms(Iterable<? extends Extension.ExtensionPlatformRelease> elements) {
      for (Extension.ExtensionPlatformRelease element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutableExtensionRelease ImmutableExtensionRelease}.
     * @return An immutable instance of ExtensionRelease
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableExtensionRelease build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableExtensionRelease(release, createUnmodifiableSet(platforms));
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_RELEASE) != 0) attributes.add("release");
      return "Cannot build ExtensionRelease, some of required attributes are not set " + attributes;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  /** Unmodifiable set constructed from list to avoid rehashing. */
  private static <T> Set<T> createUnmodifiableSet(List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptySet();
    case 1: return Collections.singleton(list.get(0));
    default:
      Set<T> set = new LinkedHashSet<>(list.size());
      set.addAll(list);
      return Collections.unmodifiableSet(set);
    }
  }
}
