package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * Immutable implementation of {@link Platform}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutablePlatform.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutablePlatform implements Platform {
  private final ArtifactKey id;
  private final Set<Release> releases;

  private ImmutablePlatform(ArtifactKey id, Set<Release> releases) {
    this.id = id;
    this.releases = releases;
  }

  /**
   * @return The value of the {@code id} attribute
   */
  @JsonProperty("id")
  @JsonUnwrapped
  @Override
  public ArtifactKey getId() {
    return id;
  }

  /**
   * @return The value of the {@code releases} attribute
   */
  @JsonProperty("releases")
  @Override
  public Set<Release> getReleases() {
    return releases;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Platform#getId() id} attribute.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for id
   * @return A modified copy of the {@code this} object
   */
  public final ImmutablePlatform withId(ArtifactKey value) {
    if (this.id == value) return this;
    ArtifactKey newValue = Objects.requireNonNull(value, "id");
    return new ImmutablePlatform(newValue, this.releases);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Platform#getReleases() releases}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutablePlatform withReleases(Release... elements) {
    Set<Release> newValue = createUnmodifiableSet(createSafeList(Arrays.asList(elements), true, false));
    return new ImmutablePlatform(this.id, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Platform#getReleases() releases}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of releases elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutablePlatform withReleases(Iterable<? extends Release> elements) {
    if (this.releases == elements) return this;
    Set<Release> newValue = createUnmodifiableSet(createSafeList(elements, true, false));
    return new ImmutablePlatform(this.id, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutablePlatform} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutablePlatform
        && equalTo((ImmutablePlatform) another);
  }

  private boolean equalTo(ImmutablePlatform another) {
    return id.equals(another.id);
  }

  /**
   * Computes a hash code from attributes: {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + id.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Platform} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Platform{"
        + "id=" + id
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Platform {
    ArtifactKey id;
    Set<Release> releases = Collections.emptySet();
    @JsonProperty("id")
    @JsonUnwrapped
    public void setId(ArtifactKey id) {
      this.id = id;
    }
    @JsonProperty("releases")
    public void setReleases(Set<Release> releases) {
      this.releases = releases;
    }
    @Override
    public ArtifactKey getId() { throw new UnsupportedOperationException(); }
    @Override
    public Set<Release> getReleases() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutablePlatform fromJson(Json json) {
    ImmutablePlatform.Builder builder = ImmutablePlatform.builder();
    if (json.id != null) {
      builder.id(json.id);
    }
    if (json.releases != null) {
      builder.addAllReleases(json.releases);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Platform} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Platform instance
   */
  public static ImmutablePlatform copyOf(Platform instance) {
    if (instance instanceof ImmutablePlatform) {
      return (ImmutablePlatform) instance;
    }
    return ImmutablePlatform.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutablePlatform ImmutablePlatform}.
   * <pre>
   * ImmutablePlatform.builder()
   *    .id(io.quarkus.registry.model.ArtifactKey) // required {@link Platform#getId() id}
   *    .addReleases|addAllReleases(io.quarkus.registry.model.Release) // {@link Platform#getReleases() releases} elements
   *    .build();
   * </pre>
   * @return A new ImmutablePlatform builder
   */
  public static ImmutablePlatform.Builder builder() {
    return new ImmutablePlatform.Builder();
  }

  /**
   * Builds instances of type {@link ImmutablePlatform ImmutablePlatform}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_ID = 0x1L;
    private long initBits = 0x1L;

    private ArtifactKey id;
    private List<Release> releases = new ArrayList<Release>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code ModifiablePlatform} instance.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(ModifiablePlatform instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance.idIsSet()) {
        id(instance.getId());
      }
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Fill a builder with attribute values from the provided {@code Platform} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Platform instance) {
      Objects.requireNonNull(instance, "instance");
      if (instance instanceof ModifiablePlatform) {
        return from((ModifiablePlatform) instance);
      }
      id(instance.getId());
      addAllReleases(instance.getReleases());
      return this;
    }

    /**
     * Initializes the value for the {@link Platform#getId() id} attribute.
     * @param id The value for id 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("id")
    @JsonUnwrapped
    public final Builder id(ArtifactKey id) {
      this.id = Objects.requireNonNull(id, "id");
      initBits &= ~INIT_BIT_ID;
      return this;
    }

    /**
     * Adds one element to {@link Platform#getReleases() releases} set.
     * @param element A releases element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release element) {
      this.releases.add(Objects.requireNonNull(element, "releases element"));
      return this;
    }

    /**
     * Adds elements to {@link Platform#getReleases() releases} set.
     * @param elements An array of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addReleases(Release... elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Platform#getReleases() releases} set.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("releases")
    public final Builder releases(Iterable<? extends Release> elements) {
      this.releases.clear();
      return addAllReleases(elements);
    }

    /**
     * Adds elements to {@link Platform#getReleases() releases} set.
     * @param elements An iterable of releases elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllReleases(Iterable<? extends Release> elements) {
      for (Release element : elements) {
        this.releases.add(Objects.requireNonNull(element, "releases element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutablePlatform ImmutablePlatform}.
     * @return An immutable instance of Platform
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutablePlatform build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutablePlatform(id, createUnmodifiableSet(releases));
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_ID) != 0) attributes.add("id");
      return "Cannot build Platform, some of required attributes are not set " + attributes;
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  /** Unmodifiable set constructed from list to avoid rehashing. */
  private static <T> Set<T> createUnmodifiableSet(List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptySet();
    case 1: return Collections.singleton(list.get(0));
    default:
      Set<T> set = new LinkedHashSet<>(list.size());
      set.addAll(list);
      return Collections.unmodifiableSet(set);
    }
  }
}
