package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import io.quarkus.dependencies.Category;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Objects;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import org.apache.maven.artifact.versioning.ComparableVersion;

/**
 * Immutable implementation of {@link Registry}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableRegistry.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableRegistry implements Registry {
  private final SortedMap<ComparableVersion, Map<String, String>> coreVersions;
  private final Set<Extension> extensions;
  private final Set<Platform> platforms;
  private final Set<Category> categories;

  private ImmutableRegistry(
      SortedMap<ComparableVersion, Map<String, String>> coreVersions,
      Set<Extension> extensions,
      Set<Platform> platforms,
      Set<Category> categories) {
    this.coreVersions = coreVersions;
    this.extensions = extensions;
    this.platforms = platforms;
    this.categories = categories;
  }

  /**
   * @return The value of the {@code coreVersions} attribute
   */
  @JsonProperty("core-versions")
  @Override
  public SortedMap<ComparableVersion, Map<String, String>> getCoreVersions() {
    return coreVersions;
  }

  /**
   * @return The value of the {@code extensions} attribute
   */
  @JsonProperty("extensions")
  @Override
  public Set<Extension> getExtensions() {
    return extensions;
  }

  /**
   * @return The value of the {@code platforms} attribute
   */
  @JsonProperty("platforms")
  @Override
  public Set<Platform> getPlatforms() {
    return platforms;
  }

  /**
   * @return The value of the {@code categories} attribute
   */
  @JsonProperty("categories")
  @Override
  public Set<Category> getCategories() {
    return categories;
  }

  /**
   * Copy the current immutable object by replacing the {@link Registry#getCoreVersions() coreVersions} map with the specified map.
   * Nulls are not permitted as keys or values.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param entries The entries to be added to the coreVersions map
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withCoreVersions(Map<? extends ComparableVersion, ? extends Map<String, String>> entries) {
    if (this.coreVersions == entries) return this;
    SortedMap<ComparableVersion, Map<String, String>> newValue = createUnmodifiableSortedMap(true, true, false, entries);
    return new ImmutableRegistry(newValue, this.extensions, this.platforms, this.categories);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getExtensions() extensions}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withExtensions(Extension... elements) {
    Set<Extension> newValue = createUnmodifiableSet(createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableRegistry(this.coreVersions, newValue, this.platforms, this.categories);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getExtensions() extensions}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of extensions elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withExtensions(Iterable<? extends Extension> elements) {
    if (this.extensions == elements) return this;
    Set<Extension> newValue = createUnmodifiableSet(createSafeList(elements, true, false));
    return new ImmutableRegistry(this.coreVersions, newValue, this.platforms, this.categories);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getPlatforms() platforms}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withPlatforms(Platform... elements) {
    Set<Platform> newValue = createUnmodifiableSet(createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableRegistry(this.coreVersions, this.extensions, newValue, this.categories);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getPlatforms() platforms}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of platforms elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withPlatforms(Iterable<? extends Platform> elements) {
    if (this.platforms == elements) return this;
    Set<Platform> newValue = createUnmodifiableSet(createSafeList(elements, true, false));
    return new ImmutableRegistry(this.coreVersions, this.extensions, newValue, this.categories);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getCategories() categories}.
   * @param elements The elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withCategories(Category... elements) {
    Set<Category> newValue = createUnmodifiableSet(createSafeList(Arrays.asList(elements), true, false));
    return new ImmutableRegistry(this.coreVersions, this.extensions, this.platforms, newValue);
  }

  /**
   * Copy the current immutable object with elements that replace the content of {@link Registry#getCategories() categories}.
   * A shallow reference equality check is used to prevent copying of the same value by returning {@code this}.
   * @param elements An iterable of categories elements to set
   * @return A modified copy of {@code this} object
   */
  public final ImmutableRegistry withCategories(Iterable<? extends Category> elements) {
    if (this.categories == elements) return this;
    Set<Category> newValue = createUnmodifiableSet(createSafeList(elements, true, false));
    return new ImmutableRegistry(this.coreVersions, this.extensions, this.platforms, newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableRegistry} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableRegistry
        && equalTo((ImmutableRegistry) another);
  }

  private boolean equalTo(ImmutableRegistry another) {
    return coreVersions.equals(another.coreVersions)
        && extensions.equals(another.extensions)
        && platforms.equals(another.platforms)
        && categories.equals(another.categories);
  }

  /**
   * Computes a hash code from attributes: {@code coreVersions}, {@code extensions}, {@code platforms}, {@code categories}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + coreVersions.hashCode();
    h += (h << 5) + extensions.hashCode();
    h += (h << 5) + platforms.hashCode();
    h += (h << 5) + categories.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Registry} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Registry{"
        + "coreVersions=" + coreVersions
        + ", extensions=" + extensions
        + ", platforms=" + platforms
        + ", categories=" + categories
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Registry {
    SortedMap<ComparableVersion, Map<String, String>> coreVersions = createUnmodifiableSortedMap(true, false, false, Collections.<ComparableVersion, Map<String, String>>emptyMap());
    Set<Extension> extensions = Collections.emptySet();
    Set<Platform> platforms = Collections.emptySet();
    Set<Category> categories = Collections.emptySet();
    @JsonProperty("core-versions")
    public void setCoreVersions(SortedMap<ComparableVersion, Map<String, String>> coreVersions) {
      this.coreVersions = coreVersions;
    }
    @JsonProperty("extensions")
    public void setExtensions(Set<Extension> extensions) {
      this.extensions = extensions;
    }
    @JsonProperty("platforms")
    public void setPlatforms(Set<Platform> platforms) {
      this.platforms = platforms;
    }
    @JsonProperty("categories")
    public void setCategories(Set<Category> categories) {
      this.categories = categories;
    }
    @Override
    public SortedMap<ComparableVersion, Map<String, String>> getCoreVersions() { throw new UnsupportedOperationException(); }
    @Override
    public Set<Extension> getExtensions() { throw new UnsupportedOperationException(); }
    @Override
    public Set<Platform> getPlatforms() { throw new UnsupportedOperationException(); }
    @Override
    public Set<Category> getCategories() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableRegistry fromJson(Json json) {
    ImmutableRegistry.Builder builder = ImmutableRegistry.builder();
    if (json.coreVersions != null) {
      builder.putAllCoreVersions(json.coreVersions);
    }
    if (json.extensions != null) {
      builder.addAllExtensions(json.extensions);
    }
    if (json.platforms != null) {
      builder.addAllPlatforms(json.platforms);
    }
    if (json.categories != null) {
      builder.addAllCategories(json.categories);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Registry} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Registry instance
   */
  public static ImmutableRegistry copyOf(Registry instance) {
    if (instance instanceof ImmutableRegistry) {
      return (ImmutableRegistry) instance;
    }
    return ImmutableRegistry.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableRegistry ImmutableRegistry}.
   * <pre>
   * ImmutableRegistry.builder()
   *    .putCoreVersions|putAllCoreVersions(org.apache.maven.artifact.versioning.ComparableVersion =&gt; Map&amp;lt;String, String&amp;gt;) // {@link Registry#getCoreVersions() coreVersions} mappings
   *    .addExtensions|addAllExtensions(io.quarkus.registry.model.Extension) // {@link Registry#getExtensions() extensions} elements
   *    .addPlatforms|addAllPlatforms(io.quarkus.registry.model.Platform) // {@link Registry#getPlatforms() platforms} elements
   *    .addCategories|addAllCategories(io.quarkus.dependencies.Category) // {@link Registry#getCategories() categories} elements
   *    .build();
   * </pre>
   * @return A new ImmutableRegistry builder
   */
  public static ImmutableRegistry.Builder builder() {
    return new ImmutableRegistry.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableRegistry ImmutableRegistry}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private Map<ComparableVersion, Map<String, String>> coreVersions = new LinkedHashMap<ComparableVersion, Map<String, String>>();
    private List<Extension> extensions = new ArrayList<Extension>();
    private List<Platform> platforms = new ArrayList<Platform>();
    private List<Category> categories = new ArrayList<Category>();

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Registry} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * Collection elements and entries will be added, not replaced.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Registry instance) {
      Objects.requireNonNull(instance, "instance");
      putAllCoreVersions(instance.getCoreVersions());
      addAllExtensions(instance.getExtensions());
      addAllPlatforms(instance.getPlatforms());
      addAllCategories(instance.getCategories());
      return this;
    }

    /**
     * Put one entry to the {@link Registry#getCoreVersions() coreVersions} map.
     * @param key The key in the coreVersions map
     * @param value The associated value in the coreVersions map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putCoreVersions(ComparableVersion key, Map<String, String> value) {
      this.coreVersions.put(
          Objects.requireNonNull(key, "coreVersions key"),
          Objects.requireNonNull(value, "coreVersions value"));
      return this;
    }

    /**
     * Put one entry to the {@link Registry#getCoreVersions() coreVersions} map. Nulls are not permitted
     * @param entry The key and value entry
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putCoreVersions(Map.Entry<? extends ComparableVersion, ? extends Map<String, String>> entry) {
      ComparableVersion k = entry.getKey();
      Map<String, String> v = entry.getValue();
      this.coreVersions.put(
          Objects.requireNonNull(k, "coreVersions key"),
          Objects.requireNonNull(v, "coreVersions value"));
      return this;
    }

    /**
     * Sets or replaces all mappings from the specified map as entries for the {@link Registry#getCoreVersions() coreVersions} map. Nulls are not permitted
     * @param entries The entries that will be added to the coreVersions map
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("core-versions")
    public final Builder coreVersions(Map<? extends ComparableVersion, ? extends Map<String, String>> entries) {
      this.coreVersions.clear();
      return putAllCoreVersions(entries);
    }

    /**
     * Put all mappings from the specified map as entries to {@link Registry#getCoreVersions() coreVersions} map. Nulls are not permitted
     * @param entries The entries that will be added to the coreVersions map
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder putAllCoreVersions(Map<? extends ComparableVersion, ? extends Map<String, String>> entries) {
      for (Map.Entry<? extends ComparableVersion, ? extends Map<String, String>> e : entries.entrySet()) {
        ComparableVersion k = e.getKey();
        Map<String, String> v = e.getValue();
        this.coreVersions.put(
            Objects.requireNonNull(k, "coreVersions key"),
            Objects.requireNonNull(v, "coreVersions value"));
      }
      return this;
    }

    /**
     * Adds one element to {@link Registry#getExtensions() extensions} set.
     * @param element A extensions element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addExtensions(Extension element) {
      this.extensions.add(Objects.requireNonNull(element, "extensions element"));
      return this;
    }

    /**
     * Adds elements to {@link Registry#getExtensions() extensions} set.
     * @param elements An array of extensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addExtensions(Extension... elements) {
      for (Extension element : elements) {
        this.extensions.add(Objects.requireNonNull(element, "extensions element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Registry#getExtensions() extensions} set.
     * @param elements An iterable of extensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("extensions")
    public final Builder extensions(Iterable<? extends Extension> elements) {
      this.extensions.clear();
      return addAllExtensions(elements);
    }

    /**
     * Adds elements to {@link Registry#getExtensions() extensions} set.
     * @param elements An iterable of extensions elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllExtensions(Iterable<? extends Extension> elements) {
      for (Extension element : elements) {
        this.extensions.add(Objects.requireNonNull(element, "extensions element"));
      }
      return this;
    }

    /**
     * Adds one element to {@link Registry#getPlatforms() platforms} set.
     * @param element A platforms element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Platform element) {
      this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      return this;
    }

    /**
     * Adds elements to {@link Registry#getPlatforms() platforms} set.
     * @param elements An array of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addPlatforms(Platform... elements) {
      for (Platform element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Registry#getPlatforms() platforms} set.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("platforms")
    public final Builder platforms(Iterable<? extends Platform> elements) {
      this.platforms.clear();
      return addAllPlatforms(elements);
    }

    /**
     * Adds elements to {@link Registry#getPlatforms() platforms} set.
     * @param elements An iterable of platforms elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllPlatforms(Iterable<? extends Platform> elements) {
      for (Platform element : elements) {
        this.platforms.add(Objects.requireNonNull(element, "platforms element"));
      }
      return this;
    }

    /**
     * Adds one element to {@link Registry#getCategories() categories} set.
     * @param element A categories element
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addCategories(Category element) {
      this.categories.add(Objects.requireNonNull(element, "categories element"));
      return this;
    }

    /**
     * Adds elements to {@link Registry#getCategories() categories} set.
     * @param elements An array of categories elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addCategories(Category... elements) {
      for (Category element : elements) {
        this.categories.add(Objects.requireNonNull(element, "categories element"));
      }
      return this;
    }


    /**
     * Sets or replaces all elements for {@link Registry#getCategories() categories} set.
     * @param elements An iterable of categories elements
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("categories")
    public final Builder categories(Iterable<? extends Category> elements) {
      this.categories.clear();
      return addAllCategories(elements);
    }

    /**
     * Adds elements to {@link Registry#getCategories() categories} set.
     * @param elements An iterable of categories elements
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder addAllCategories(Iterable<? extends Category> elements) {
      for (Category element : elements) {
        this.categories.add(Objects.requireNonNull(element, "categories element"));
      }
      return this;
    }

    /**
     * Builds a new {@link ImmutableRegistry ImmutableRegistry}.
     * @return An immutable instance of Registry
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableRegistry build() {
      return new ImmutableRegistry(
          createUnmodifiableSortedMap(true, false, false, coreVersions),
          createUnmodifiableSet(extensions),
          createUnmodifiableSet(platforms),
          createUnmodifiableSet(categories));
    }
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  /** Unmodifiable set constructed from list to avoid rehashing. */
  private static <T> Set<T> createUnmodifiableSet(List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptySet();
    case 1: return Collections.singleton(list.get(0));
    default:
      Set<T> set = new LinkedHashSet<>(list.size());
      set.addAll(list);
      return Collections.unmodifiableSet(set);
    }
  }

  private static <K extends Comparable<K>, V> NavigableMap<K, V> createUnmodifiableSortedMap(boolean reverse, boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    TreeMap<K, V> sortedMap = reverse
        ? new TreeMap<>(Collections.reverseOrder())
        : new TreeMap<>();

    if (checkNulls || skipNulls) {
      for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
        K k = e.getKey();
        V v = e.getValue();
        if (skipNulls) {
          if (k == null || v == null) continue;
        } else if (checkNulls) {
          Objects.requireNonNull(k, "key");
          Objects.requireNonNull(v, "value");
        }
        sortedMap.put(k, v);
      }
    } else {
      sortedMap.putAll(map);
    }
    return Collections.unmodifiableNavigableMap(sortedMap);
  }
}
