package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link Release}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableRelease.builder()}.
 */
@SuppressWarnings({"all"})
public final class ImmutableRelease implements Release {
  private final String version;
  private final @Nullable String quarkusCore;
  private final @Nullable String repositoryURL;

  private ImmutableRelease(
      String version,
      @Nullable String quarkusCore,
      @Nullable String repositoryURL) {
    this.version = version;
    this.quarkusCore = quarkusCore;
    this.repositoryURL = repositoryURL;
  }

  /**
   * @return The value of the {@code version} attribute
   */
  @JsonProperty("version")
  @Override
  public String getVersion() {
    return version;
  }

  /**
   * @return The value of the {@code quarkusCore} attribute
   */
  @JsonProperty("quarkus-core")
  @Override
  public @Nullable String getQuarkusCore() {
    return quarkusCore;
  }

  /**
   * @return The value of the {@code repositoryURL} attribute
   */
  @JsonProperty("repository-url")
  @Override
  public @Nullable String getRepositoryURL() {
    return repositoryURL;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Release#getVersion() version} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for version
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableRelease withVersion(String value) {
    String newValue = Objects.requireNonNull(value, "version");
    if (this.version.equals(newValue)) return this;
    return new ImmutableRelease(newValue, this.quarkusCore, this.repositoryURL);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Release#getQuarkusCore() quarkusCore} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for quarkusCore (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableRelease withQuarkusCore(@Nullable String value) {
    if (Objects.equals(this.quarkusCore, value)) return this;
    return new ImmutableRelease(this.version, value, this.repositoryURL);
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Release#getRepositoryURL() repositoryURL} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for repositoryURL (can be {@code null})
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableRelease withRepositoryURL(@Nullable String value) {
    if (Objects.equals(this.repositoryURL, value)) return this;
    return new ImmutableRelease(this.version, this.quarkusCore, value);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableRelease} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableRelease
        && equalTo((ImmutableRelease) another);
  }

  private boolean equalTo(ImmutableRelease another) {
    return version.equals(another.version)
        && Objects.equals(quarkusCore, another.quarkusCore);
  }

  /**
   * Computes a hash code from attributes: {@code version}, {@code quarkusCore}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + version.hashCode();
    h += (h << 5) + Objects.hashCode(quarkusCore);
    return h;
  }

  /**
   * Prints the immutable value {@code Release} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Release{"
        + "version=" + version
        + ", quarkusCore=" + quarkusCore
        + "}";
  }

  /**
   * Utility type used to correctly read immutable object from JSON representation.
   * @deprecated Do not use this type directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonDeserialize
  @JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE)
  static final class Json implements Release {
    String version;
    String quarkusCore;
    String repositoryURL;
    @JsonProperty("version")
    public void setVersion(String version) {
      this.version = version;
    }
    @JsonProperty("quarkus-core")
    public void setQuarkusCore(@Nullable String quarkusCore) {
      this.quarkusCore = quarkusCore;
    }
    @JsonProperty("repository-url")
    public void setRepositoryURL(@Nullable String repositoryURL) {
      this.repositoryURL = repositoryURL;
    }
    @Override
    public String getVersion() { throw new UnsupportedOperationException(); }
    @Override
    public String getQuarkusCore() { throw new UnsupportedOperationException(); }
    @Override
    public String getRepositoryURL() { throw new UnsupportedOperationException(); }
  }

  /**
   * @param json A JSON-bindable data structure
   * @return An immutable value type
   * @deprecated Do not use this method directly, it exists only for the <em>Jackson</em>-binding infrastructure
   */
  @Deprecated
  @JsonCreator(mode = JsonCreator.Mode.DELEGATING)
  static ImmutableRelease fromJson(Json json) {
    ImmutableRelease.Builder builder = ImmutableRelease.builder();
    if (json.version != null) {
      builder.version(json.version);
    }
    if (json.quarkusCore != null) {
      builder.quarkusCore(json.quarkusCore);
    }
    if (json.repositoryURL != null) {
      builder.repositoryURL(json.repositoryURL);
    }
    return builder.build();
  }

  /**
   * Creates an immutable copy of a {@link Release} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Release instance
   */
  public static ImmutableRelease copyOf(Release instance) {
    if (instance instanceof ImmutableRelease) {
      return (ImmutableRelease) instance;
    }
    return ImmutableRelease.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableRelease ImmutableRelease}.
   * <pre>
   * ImmutableRelease.builder()
   *    .version(String) // required {@link Release#getVersion() version}
   *    .quarkusCore(String | null) // nullable {@link Release#getQuarkusCore() quarkusCore}
   *    .repositoryURL(String | null) // nullable {@link Release#getRepositoryURL() repositoryURL}
   *    .build();
   * </pre>
   * @return A new ImmutableRelease builder
   */
  public static ImmutableRelease.Builder builder() {
    return new ImmutableRelease.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableRelease ImmutableRelease}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  public static final class Builder {
    private static final long INIT_BIT_VERSION = 0x1L;
    private long initBits = 0x1L;

    private String version;
    private String quarkusCore;
    private String repositoryURL;

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Release} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Release instance) {
      Objects.requireNonNull(instance, "instance");
      version(instance.getVersion());
      @Nullable String quarkusCoreValue = instance.getQuarkusCore();
      if (quarkusCoreValue != null) {
        quarkusCore(quarkusCoreValue);
      }
      @Nullable String repositoryURLValue = instance.getRepositoryURL();
      if (repositoryURLValue != null) {
        repositoryURL(repositoryURLValue);
      }
      return this;
    }

    /**
     * Initializes the value for the {@link Release#getVersion() version} attribute.
     * @param version The value for version 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("version")
    public final Builder version(String version) {
      this.version = Objects.requireNonNull(version, "version");
      initBits &= ~INIT_BIT_VERSION;
      return this;
    }

    /**
     * Initializes the value for the {@link Release#getQuarkusCore() quarkusCore} attribute.
     * @param quarkusCore The value for quarkusCore (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("quarkus-core")
    public final Builder quarkusCore(@Nullable String quarkusCore) {
      this.quarkusCore = quarkusCore;
      return this;
    }

    /**
     * Initializes the value for the {@link Release#getRepositoryURL() repositoryURL} attribute.
     * @param repositoryURL The value for repositoryURL (can be {@code null})
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("repository-url")
    public final Builder repositoryURL(@Nullable String repositoryURL) {
      this.repositoryURL = repositoryURL;
      return this;
    }

    /**
     * Builds a new {@link ImmutableRelease ImmutableRelease}.
     * @return An immutable instance of Release
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableRelease build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableRelease(version, quarkusCore, repositoryURL);
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_VERSION) != 0) attributes.add("version");
      return "Cannot build Release, some of required attributes are not set " + attributes;
    }
  }
}
