package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Objects;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * A modifiable implementation of the {@link Extension Extension} type.
 * <p>Use the {@link #create()} static factory methods to create new instances.
 * Use the {@link #toImmutable()} method to convert to canonical immutable instances.
 * <p><em>ModifiableExtension is not thread-safe</em>
 * @see ImmutableExtension
 */
@SuppressWarnings({"all"})
public final class ModifiableExtension implements Extension {
  private static final long INIT_BIT_ID = 0x1L;
  private static final long INIT_BIT_NAME = 0x2L;
  private long initBits = 0x3L;

  private ArtifactKey id;
  private String name;
  private @Nullable String description;
  private final Map<String, java.lang.Object> metadata = new LinkedHashMap<String, java.lang.Object>();
  private final TreeSet<Extension.ExtensionRelease> releases = new TreeSet<Extension.ExtensionRelease>(Collections.<Extension.ExtensionRelease>reverseOrder());

  private ModifiableExtension() {}

  /**
   * Construct a modifiable instance of {@code Extension}.
   * @return A new modifiable instance
   */
  public static ModifiableExtension create() {
    return new ModifiableExtension();
  }

  /**
   * @return value of {@code id} attribute
   */
  @JsonProperty("id")
  @JsonUnwrapped
  @Override
  public final ArtifactKey getId() {
    if (!idIsSet()) {
      checkRequiredAttributes();
    }
    return id;
  }

  /**
   * @return value of {@code name} attribute
   */
  @JsonProperty("name")
  @Override
  public final String getName() {
    if (!nameIsSet()) {
      checkRequiredAttributes();
    }
    return name;
  }

  /**
   * @return value of {@code description} attribute, may be {@code null}
   */
  @JsonProperty("description")
  @Override
  public final @Nullable String getDescription() {
    return description;
  }

  /**
   * @return value of {@code metadata} attribute
   */
  @JsonProperty("metadata")
  @Override
  public final Map<String, java.lang.Object> getMetadata() {
    return metadata;
  }

  /**
   * @return modifiable sortedSet {@code releases}
   */
  @JsonProperty("releases")
  @Override
  public final SortedSet<Extension.ExtensionRelease> getReleases() {
    return releases;
  }

  /**
   * Clears the object by setting all attributes to their initial values.
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension clear() {
    initBits = 0x3L;
    id = null;
    name = null;
    description = null;
    metadata.clear();
    releases.clear();
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Extension} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension from(Extension instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance instanceof ModifiableExtension) {
      from((ModifiableExtension) instance);
      return this;
    }
    setId(instance.getId());
    setName(instance.getName());
    @Nullable String descriptionValue = instance.getDescription();
    if (descriptionValue != null) {
      setDescription(descriptionValue);
    }
    putAllMetadata(instance.getMetadata());
    addAllReleases(instance.getReleases());
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Extension} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension from(ModifiableExtension instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance.idIsSet()) {
      setId(instance.getId());
    }
    if (instance.nameIsSet()) {
      setName(instance.getName());
    }
    @Nullable String descriptionValue = instance.getDescription();
    if (descriptionValue != null) {
      setDescription(descriptionValue);
    }
    putAllMetadata(instance.getMetadata());
    addAllReleases(instance.getReleases());
    return this;
  }

  /**
   * Assigns a value to the {@link Extension#getId() id} attribute.
   * @param id The value for id
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension setId(ArtifactKey id) {
    this.id = Objects.requireNonNull(id, "id");
    initBits &= ~INIT_BIT_ID;
    return this;
  }

  /**
   * Assigns a value to the {@link Extension#getName() name} attribute.
   * @param name The value for name
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension setName(String name) {
    this.name = Objects.requireNonNull(name, "name");
    initBits &= ~INIT_BIT_NAME;
    return this;
  }

  /**
   * Assigns a value to the {@link Extension#getDescription() description} attribute.
   * @param description The value for description, can be {@code null}
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension setDescription(@Nullable String description) {
    this.description = description;
    return this;
  }

  /**
   * Put one entry to the {@link Extension#getMetadata() metadata} map.
   * @param key The key in metadata map
   * @param value The associated value in the metadata map
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension putMetadata(String key, java.lang.Object value) {
    this.metadata.put(
        Objects.requireNonNull(key, "metadata key"),
        Objects.requireNonNull(value, "metadata value"));
    return this;
  }

  /**
   * Sets or replaces all mappings from the specified map as entries for the {@link Extension#getMetadata() metadata} map.
   * Nulls are not permitted as keys or values.
   * @param entries The entries that will be added to the metadata map
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension setMetadata(Map<String, ? extends java.lang.Object> entries) {
    this.metadata.clear();
    for (Map.Entry<String, ? extends java.lang.Object> e : entries.entrySet()) {
      String k = e.getKey();
      java.lang.Object v = e.getValue();
      this.metadata.put(
          Objects.requireNonNull(k, "metadata key"),
          Objects.requireNonNull(v, "metadata value"));
    }
    return this;
  }

  /**
   * Put all mappings from the specified map as entries to the {@link Extension#getMetadata() metadata} map.
   * Nulls are not permitted as keys or values.
   * @param entries to be added to metadata map
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension putAllMetadata(Map<String, ? extends java.lang.Object> entries) {
    for (Map.Entry<String, ? extends java.lang.Object> e : entries.entrySet()) {
      String k = e.getKey();
      java.lang.Object v = e.getValue();
      this.metadata.put(
          Objects.requireNonNull(k, "metadata key"),
          Objects.requireNonNull(v, "metadata value"));
    }
    return this;
  }

  /**
   * Adds one element to {@link Extension#getReleases() releases} sortedSet.
   * @param element The releases element
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension addReleases(Extension.ExtensionRelease element) {
    Objects.requireNonNull(element, "releases element");
    this.releases.add(element);
    return this;
  }

  /**
   * Adds elements to {@link Extension#getReleases() releases} sortedSet.
   * @param elements An array of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiableExtension addReleases(Extension.ExtensionRelease... elements) {
    for (Extension.ExtensionRelease e : elements) {
      addReleases(e);
    }
    return this;
  }

  /**
   * Sets or replaces all elements for {@link Extension#getReleases() releases} sortedSet.
   * @param elements An iterable of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension setReleases(Iterable<? extends Extension.ExtensionRelease> elements) {
    this.releases.clear();
    addAllReleases(elements);
    return this;
  }

  /**
   * Adds elements to {@link Extension#getReleases() releases} sortedSet.
   * @param elements An iterable of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtension addAllReleases(Iterable<? extends Extension.ExtensionRelease> elements) {
    for (Extension.ExtensionRelease e : elements) {
      addReleases(e);
    }
    return this;
  }

  /**
   * Returns {@code true} if the required attribute {@link Extension#getId() id} is set.
   * @return {@code true} if set
   */
  public final boolean idIsSet() {
    return (initBits & INIT_BIT_ID) == 0;
  }

  /**
   * Returns {@code true} if the required attribute {@link Extension#getName() name} is set.
   * @return {@code true} if set
   */
  public final boolean nameIsSet() {
    return (initBits & INIT_BIT_NAME) == 0;
  }


  /**
   * Reset an attribute to its initial value.
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiableExtension unsetId() {
    initBits |= INIT_BIT_ID;
    id = null;
    return this;
  }

  /**
   * Reset an attribute to its initial value.
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiableExtension unsetName() {
    initBits |= INIT_BIT_NAME;
    name = null;
    return this;
  }

  /**
   * Returns {@code true} if all required attributes are set, indicating that the object is initialized.
   * @return {@code true} if set
   */
  public final boolean isInitialized() {
    return initBits == 0;
  }

  private void checkRequiredAttributes() {
    if (!isInitialized()) {
      throw new IllegalStateException(formatRequiredAttributesMessage());
    }
  }

  private String formatRequiredAttributesMessage() {
    List<String> attributes = new ArrayList<>();
    if (!idIsSet()) attributes.add("id");
    if (!nameIsSet()) attributes.add("name");
    return "Extension is not initialized, some of the required attributes are not set " + attributes;
  }

  /**
   * Converts to {@link ImmutableExtension ImmutableExtension}.
   * @return An immutable instance of Extension
   */
  public final ImmutableExtension toImmutable() {
    checkRequiredAttributes();
    return ImmutableExtension.copyOf(this);
  }

  /**
   * This instance is equal to all instances of {@code ModifiableExtension} that have equal attribute values.
   * An uninitialized instance is equal only to itself.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    if (!(another instanceof ModifiableExtension)) return false;
    ModifiableExtension other = (ModifiableExtension) another;
    if (!isInitialized() || !other.isInitialized()) {
      return false;
    }
    return equalTo(other);
  }

  private boolean equalTo(ModifiableExtension another) {
    return id.equals(another.id);
  }

  /**
   * Computes a hash code from attributes: {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + id.hashCode();
    return h;
  }

  /**
   * Generates a string representation of this {@code Extension}.
   * If uninitialized, some attribute values may appear as question marks.
   * @return A string representation
   */
  @Override
  public String toString() {
    return "ModifiableExtension{"
        + "id="  + (idIsSet() ? getId() : "?")
        + "}";
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  private static <T extends Comparable<T>> NavigableSet<T> createUnmodifiableSortedSet(boolean reverse, List<T> list) {
    TreeSet<T> set = reverse
        ? new TreeSet<>(Collections.reverseOrder())
        : new TreeSet<>();
    set.addAll(list);
    return Collections.unmodifiableNavigableSet(set);
  }

  private static <K, V> Map<K, V> createUnmodifiableMap(boolean checkNulls, boolean skipNulls, Map<? extends K, ? extends V> map) {
    switch (map.size()) {
    case 0: return Collections.emptyMap();
    case 1: {
      Map.Entry<? extends K, ? extends V> e = map.entrySet().iterator().next();
      K k = e.getKey();
      V v = e.getValue();
      if (checkNulls) {
        Objects.requireNonNull(k, "key");
        Objects.requireNonNull(v, "value");
      }
      if (skipNulls && (k == null || v == null)) {
        return Collections.emptyMap();
      }
      return Collections.singletonMap(k, v);
    }
    default: {
      Map<K, V> linkedMap = new LinkedHashMap<>(map.size());
      if (skipNulls || checkNulls) {
        for (Map.Entry<? extends K, ? extends V> e : map.entrySet()) {
          K k = e.getKey();
          V v = e.getValue();
          if (skipNulls) {
            if (k == null || v == null) continue;
          } else if (checkNulls) {
            Objects.requireNonNull(k, "key");
            Objects.requireNonNull(v, "value");
          }
          linkedMap.put(k, v);
        }
      } else {
        linkedMap.putAll(map);
      }
      return Collections.unmodifiableMap(linkedMap);
    }
    }
  }
}
