package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * A modifiable implementation of the {@link Extension.ExtensionRelease ExtensionRelease} type.
 * <p>Use the {@link #create()} static factory methods to create new instances.
 * Use the {@link #toImmutable()} method to convert to canonical immutable instances.
 * <p><em>ModifiableExtensionRelease is not thread-safe</em>
 * @see ImmutableExtensionRelease
 */
@SuppressWarnings({"all"})
public final class ModifiableExtensionRelease implements Extension.ExtensionRelease {
  private static final long INIT_BIT_RELEASE = 0x1L;
  private long initBits = 0x1L;

  private Release release;
  private final LinkedHashSet<Extension.ExtensionPlatformRelease> platforms = new LinkedHashSet<Extension.ExtensionPlatformRelease>();

  private ModifiableExtensionRelease() {}

  /**
   * Construct a modifiable instance of {@code ExtensionRelease}.
   * @return A new modifiable instance
   */
  public static ModifiableExtensionRelease create() {
    return new ModifiableExtensionRelease();
  }

  /**
   * @return value of {@code release} attribute
   */
  @JsonProperty("release")
  @JsonUnwrapped
  @Override
  public final Release getRelease() {
    if (!releaseIsSet()) {
      checkRequiredAttributes();
    }
    return release;
  }

  /**
   * @return modifiable set {@code platforms}
   */
  @JsonProperty("platforms")
  @Override
  public final Set<Extension.ExtensionPlatformRelease> getPlatforms() {
    return platforms;
  }

  /**
   * Clears the object by setting all attributes to their initial values.
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease clear() {
    initBits = 0x1L;
    release = null;
    platforms.clear();
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Extension.ExtensionRelease} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease from(Extension.ExtensionRelease instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance instanceof ModifiableExtensionRelease) {
      from((ModifiableExtensionRelease) instance);
      return this;
    }
    setRelease(instance.getRelease());
    addAllPlatforms(instance.getPlatforms());
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Extension.ExtensionRelease} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease from(ModifiableExtensionRelease instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance.releaseIsSet()) {
      setRelease(instance.getRelease());
    }
    addAllPlatforms(instance.getPlatforms());
    return this;
  }

  /**
   * Assigns a value to the {@link Extension.ExtensionRelease#getRelease() release} attribute.
   * @param release The value for release
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease setRelease(Release release) {
    this.release = Objects.requireNonNull(release, "release");
    initBits &= ~INIT_BIT_RELEASE;
    return this;
  }

  /**
   * Adds one element to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
   * @param element The platforms element
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease addPlatforms(Extension.ExtensionPlatformRelease element) {
    Objects.requireNonNull(element, "platforms element");
    this.platforms.add(element);
    return this;
  }

  /**
   * Adds elements to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
   * @param elements An array of platforms elements
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiableExtensionRelease addPlatforms(Extension.ExtensionPlatformRelease... elements) {
    for (Extension.ExtensionPlatformRelease e : elements) {
      addPlatforms(e);
    }
    return this;
  }

  /**
   * Sets or replaces all elements for {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
   * @param elements An iterable of platforms elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease setPlatforms(Iterable<? extends Extension.ExtensionPlatformRelease> elements) {
    this.platforms.clear();
    addAllPlatforms(elements);
    return this;
  }

  /**
   * Adds elements to {@link Extension.ExtensionRelease#getPlatforms() platforms} set.
   * @param elements An iterable of platforms elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiableExtensionRelease addAllPlatforms(Iterable<? extends Extension.ExtensionPlatformRelease> elements) {
    for (Extension.ExtensionPlatformRelease e : elements) {
      addPlatforms(e);
    }
    return this;
  }

  /**
   * Returns {@code true} if the required attribute {@link Extension.ExtensionRelease#getRelease() release} is set.
   * @return {@code true} if set
   */
  public final boolean releaseIsSet() {
    return (initBits & INIT_BIT_RELEASE) == 0;
  }


  /**
   * Reset an attribute to its initial value.
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiableExtensionRelease unsetRelease() {
    initBits |= INIT_BIT_RELEASE;
    release = null;
    return this;
  }

  /**
   * Returns {@code true} if all required attributes are set, indicating that the object is initialized.
   * @return {@code true} if set
   */
  public final boolean isInitialized() {
    return initBits == 0;
  }

  private void checkRequiredAttributes() {
    if (!isInitialized()) {
      throw new IllegalStateException(formatRequiredAttributesMessage());
    }
  }

  private String formatRequiredAttributesMessage() {
    List<String> attributes = new ArrayList<>();
    if (!releaseIsSet()) attributes.add("release");
    return "ExtensionRelease is not initialized, some of the required attributes are not set " + attributes;
  }

  /**
   * Converts to {@link ImmutableExtensionRelease ImmutableExtensionRelease}.
   * @return An immutable instance of ExtensionRelease
   */
  public final ImmutableExtensionRelease toImmutable() {
    checkRequiredAttributes();
    return ImmutableExtensionRelease.copyOf(this);
  }

  /**
   * This instance is equal to all instances of {@code ModifiableExtensionRelease} that have equal attribute values.
   * An uninitialized instance is equal only to itself.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    if (!(another instanceof ModifiableExtensionRelease)) return false;
    ModifiableExtensionRelease other = (ModifiableExtensionRelease) another;
    if (!isInitialized() || !other.isInitialized()) {
      return false;
    }
    return equalTo(other);
  }

  private boolean equalTo(ModifiableExtensionRelease another) {
    return release.equals(another.release);
  }

  /**
   * Computes a hash code from attributes: {@code release}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + release.hashCode();
    return h;
  }

  /**
   * Generates a string representation of this {@code ExtensionRelease}.
   * If uninitialized, some attribute values may appear as question marks.
   * @return A string representation
   */
  @Override
  public String toString() {
    return "ModifiableExtensionRelease{"
        + "release="  + (releaseIsSet() ? getRelease() : "?")
        + "}";
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  /** Unmodifiable set constructed from list to avoid rehashing. */
  private static <T> Set<T> createUnmodifiableSet(List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptySet();
    case 1: return Collections.singleton(list.get(0));
    default:
      Set<T> set = new LinkedHashSet<>(list.size());
      set.addAll(list);
      return Collections.unmodifiableSet(set);
    }
  }
}
