package io.quarkus.registry.model;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonUnwrapped;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * A modifiable implementation of the {@link Platform Platform} type.
 * <p>Use the {@link #create()} static factory methods to create new instances.
 * Use the {@link #toImmutable()} method to convert to canonical immutable instances.
 * <p><em>ModifiablePlatform is not thread-safe</em>
 * @see ImmutablePlatform
 */
@SuppressWarnings({"all"})
public final class ModifiablePlatform implements Platform {
  private static final long INIT_BIT_ID = 0x1L;
  private long initBits = 0x1L;

  private ArtifactKey id;
  private final LinkedHashSet<Release> releases = new LinkedHashSet<Release>();

  private ModifiablePlatform() {}

  /**
   * Construct a modifiable instance of {@code Platform}.
   * @return A new modifiable instance
   */
  public static ModifiablePlatform create() {
    return new ModifiablePlatform();
  }

  /**
   * @return value of {@code id} attribute
   */
  @JsonProperty("id")
  @JsonUnwrapped
  @Override
  public final ArtifactKey getId() {
    if (!idIsSet()) {
      checkRequiredAttributes();
    }
    return id;
  }

  /**
   * @return modifiable set {@code releases}
   */
  @JsonProperty("releases")
  @Override
  public final Set<Release> getReleases() {
    return releases;
  }

  /**
   * Clears the object by setting all attributes to their initial values.
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform clear() {
    initBits = 0x1L;
    id = null;
    releases.clear();
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Platform} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform from(Platform instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance instanceof ModifiablePlatform) {
      from((ModifiablePlatform) instance);
      return this;
    }
    setId(instance.getId());
    addAllReleases(instance.getReleases());
    return this;
  }

  /**
   * Fill this modifiable instance with attribute values from the provided {@link Platform} instance.
   * Regular attribute values will be overridden, i.e. replaced with ones of an instance.
   * Any of the instance's absent optional values will not be copied (will not override current values).
   * Collection elements and entries will be added, not replaced.
   * @param instance The instance from which to copy values
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform from(ModifiablePlatform instance) {
    Objects.requireNonNull(instance, "instance");
    if (instance.idIsSet()) {
      setId(instance.getId());
    }
    addAllReleases(instance.getReleases());
    return this;
  }

  /**
   * Assigns a value to the {@link Platform#getId() id} attribute.
   * @param id The value for id
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform setId(ArtifactKey id) {
    this.id = Objects.requireNonNull(id, "id");
    initBits &= ~INIT_BIT_ID;
    return this;
  }

  /**
   * Adds one element to {@link Platform#getReleases() releases} set.
   * @param element The releases element
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform addReleases(Release element) {
    Objects.requireNonNull(element, "releases element");
    this.releases.add(element);
    return this;
  }

  /**
   * Adds elements to {@link Platform#getReleases() releases} set.
   * @param elements An array of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiablePlatform addReleases(Release... elements) {
    for (Release e : elements) {
      addReleases(e);
    }
    return this;
  }

  /**
   * Sets or replaces all elements for {@link Platform#getReleases() releases} set.
   * @param elements An iterable of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform setReleases(Iterable<? extends Release> elements) {
    this.releases.clear();
    addAllReleases(elements);
    return this;
  }

  /**
   * Adds elements to {@link Platform#getReleases() releases} set.
   * @param elements An iterable of releases elements
   * @return {@code this} for use in a chained invocation
   */
  public ModifiablePlatform addAllReleases(Iterable<? extends Release> elements) {
    for (Release e : elements) {
      addReleases(e);
    }
    return this;
  }

  /**
   * Returns {@code true} if the required attribute {@link Platform#getId() id} is set.
   * @return {@code true} if set
   */
  public final boolean idIsSet() {
    return (initBits & INIT_BIT_ID) == 0;
  }


  /**
   * Reset an attribute to its initial value.
   * @return {@code this} for use in a chained invocation
   */
  public final ModifiablePlatform unsetId() {
    initBits |= INIT_BIT_ID;
    id = null;
    return this;
  }

  /**
   * Returns {@code true} if all required attributes are set, indicating that the object is initialized.
   * @return {@code true} if set
   */
  public final boolean isInitialized() {
    return initBits == 0;
  }

  private void checkRequiredAttributes() {
    if (!isInitialized()) {
      throw new IllegalStateException(formatRequiredAttributesMessage());
    }
  }

  private String formatRequiredAttributesMessage() {
    List<String> attributes = new ArrayList<>();
    if (!idIsSet()) attributes.add("id");
    return "Platform is not initialized, some of the required attributes are not set " + attributes;
  }

  /**
   * Converts to {@link ImmutablePlatform ImmutablePlatform}.
   * @return An immutable instance of Platform
   */
  public final ImmutablePlatform toImmutable() {
    checkRequiredAttributes();
    return ImmutablePlatform.copyOf(this);
  }

  /**
   * This instance is equal to all instances of {@code ModifiablePlatform} that have equal attribute values.
   * An uninitialized instance is equal only to itself.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    if (!(another instanceof ModifiablePlatform)) return false;
    ModifiablePlatform other = (ModifiablePlatform) another;
    if (!isInitialized() || !other.isInitialized()) {
      return false;
    }
    return equalTo(other);
  }

  private boolean equalTo(ModifiablePlatform another) {
    return id.equals(another.id);
  }

  /**
   * Computes a hash code from attributes: {@code id}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + id.hashCode();
    return h;
  }

  /**
   * Generates a string representation of this {@code Platform}.
   * If uninitialized, some attribute values may appear as question marks.
   * @return A string representation
   */
  @Override
  public String toString() {
    return "ModifiablePlatform{"
        + "id="  + (idIsSet() ? getId() : "?")
        + "}";
  }

  private static <T> List<T> createSafeList(Iterable<? extends T> iterable, boolean checkNulls, boolean skipNulls) {
    ArrayList<T> list;
    if (iterable instanceof Collection<?>) {
      int size = ((Collection<?>) iterable).size();
      if (size == 0) return Collections.emptyList();
      list = new ArrayList<>();
    } else {
      list = new ArrayList<>();
    }
    for (T element : iterable) {
      if (skipNulls && element == null) continue;
      if (checkNulls) Objects.requireNonNull(element, "element");
      list.add(element);
    }
    return list;
  }

  /** Unmodifiable set constructed from list to avoid rehashing. */
  private static <T> Set<T> createUnmodifiableSet(List<T> list) {
    switch(list.size()) {
    case 0: return Collections.emptySet();
    case 1: return Collections.singleton(list.get(0));
    default:
      Set<T> set = new LinkedHashSet<>(list.size());
      set.addAll(list);
      return Collections.unmodifiableSet(set);
    }
  }
}
